"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.RouteBuilder = void 0;
const tslib_1 = require("tslib");
const lodash_1 = tslib_1.__importDefault(require("lodash"));
const _utils_1 = require("./@utils");
class RouteBuilder {
    constructor(router, sourceType, buildingParts = [], leavingGroupSet = new Set()) {
        Object.defineProperty(this, "router", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: router
        });
        Object.defineProperty(this, "sourceType", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: sourceType
        });
        Object.defineProperty(this, "buildingParts", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: buildingParts
        });
        Object.defineProperty(this, "leavingGroupSet", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: leavingGroupSet
        });
    }
    /**
     * Route of the first building part if available.
     */
    get $route() {
        const [firstPart] = this.buildingParts;
        return typeof firstPart === 'object' ? firstPart.route : undefined;
    }
    $(route, params) {
        const buildingPart = typeof route === 'string'
            ? route
            : {
                route,
                params,
            };
        return new RouteBuilder(this.router, this.sourceType, [...this.buildingParts, buildingPart], this.leavingGroupSet);
    }
    $leave(groups) {
        if (typeof groups === 'string') {
            groups = [groups];
        }
        const leavingGroupSet = new Set([...this.leavingGroupSet, ...groups]);
        return new RouteBuilder(this.router, this.sourceType, this.buildingParts, leavingGroupSet);
    }
    $ref() {
        var _a, _b;
        const router = this.router;
        const sourceType = this.sourceType;
        const leavingGroupSet = this.leavingGroupSet;
        const groupToBuildingPartMap = new Map();
        if (sourceType !== 'none') {
            for (const [group, route] of router._groupToRouteMatchMap) {
                if (group && leavingGroupSet.has(group)) {
                    continue;
                }
                const sourceRoute = sourceType === 'current' ? route : route.$next;
                if (!sourceRoute.$matched) {
                    continue;
                }
                groupToBuildingPartMap.set(group, {
                    route: sourceRoute.$rest,
                });
            }
        }
        for (const buildingPart of this.buildingParts) {
            if (typeof buildingPart === 'string') {
                const { groups, query: buildingPartQueryMap } = parseStringBuildingPart(buildingPart, router.$groups);
                for (const { name: group, path } of groups) {
                    if (group && leavingGroupSet.has(group)) {
                        continue;
                    }
                    // Preserve the route information if already exists.
                    const route = (_a = groupToBuildingPartMap.get(group)) === null || _a === void 0 ? void 0 : _a.route;
                    groupToBuildingPartMap.set(group, {
                        path,
                        route,
                        query: buildingPartQueryMap,
                    });
                }
            }
            else {
                const group = buildingPart.route.$group;
                if (group && leavingGroupSet.has(group)) {
                    continue;
                }
                groupToBuildingPartMap.set(group, buildingPart);
            }
        }
        const pathMap = new Map();
        const queryMap = new Map();
        for (const [group, buildingPart] of groupToBuildingPartMap) {
            if ('path' in buildingPart) {
                let { path, route, query: buildingPartQueryMap } = buildingPart;
                pathMap.set(group, path);
                if (route) {
                    buildingPartQueryMap = new Map([
                        ...Array.from(route._source.queryMap).map(([key, { value }]) => [key, value]),
                        ...buildingPartQueryMap,
                    ]);
                }
                for (const [key, value] of buildingPartQueryMap) {
                    if (queryMap.has(key)) {
                        continue;
                    }
                    queryMap.set(key, value);
                }
            }
            else {
                const { route, params: paramDict = {} } = buildingPart;
                const nextSegmentDict = route._pathSegments;
                const nextSegmentNames = lodash_1.default.keys(nextSegmentDict);
                const restSegmentDict = lodash_1.default.pick(((_b = router._groupToRouteMatchMap.get(group)) === null || _b === void 0 ? void 0 : _b.$rest._pathSegments) || {}, nextSegmentNames);
                const segmentDict = lodash_1.default.fromPairs(lodash_1.default.sortBy(lodash_1.default.entries(lodash_1.default.merge(restSegmentDict, nextSegmentDict)), ([key]) => nextSegmentNames.indexOf(key)));
                const queryKeyToIdMap = route._queryKeyToIdMap;
                const queryKeys = Array.from(queryKeyToIdMap.keys());
                pathMap.set(group, (0, _utils_1.buildPath)(segmentDict, lodash_1.default.omit(paramDict, queryKeys)));
                const { queryMap: sourceQueryMap } = route._source;
                for (const [key, { id, value }] of sourceQueryMap) {
                    const routeQueryId = queryKeyToIdMap.get(key);
                    if (queryMap.has(key) ||
                        routeQueryId === undefined ||
                        !(0, _utils_1.isQueryIdsMatched)(routeQueryId, id)) {
                        continue;
                    }
                    queryMap.set(key, value);
                }
                const restParamKeys = lodash_1.default.difference(Object.keys(paramDict), lodash_1.default.difference(Object.keys(segmentDict), queryKeys));
                for (const key of restParamKeys) {
                    if (!queryKeyToIdMap.has(key)) {
                        throw new Error(`Parameter "${key}" is defined as neither segment nor query`);
                    }
                    // Note a given param could be `undefined` here to remove the query.
                    // The reason why we use `undefined` instead of deleting the key is
                    // to ensure this overrides route queries iterated later.
                    queryMap.set(key, paramDict[key]);
                }
            }
        }
        return (0, _utils_1.buildRef)(pathMap, queryMap);
    }
    $href() {
        let ref = this.$ref();
        if (!ref.startsWith('/')) {
            ref = `/${ref}`;
        }
        return this.router._history.getHRefByRef(ref);
    }
    /**
     * Perform a `history.push()` with `this.$ref()`.
     */
    $push(options) {
        const ref = this.$ref();
        this.router._push(ref, options);
    }
    /**
     * Perform a `history.replace()` with `this.$ref()`.
     */
    $replace(options) {
        const ref = this.$ref();
        this.router._replace(ref, options);
    }
}
exports.RouteBuilder = RouteBuilder;
function parseStringBuildingPart(part, groups) {
    const searchIndex = part.indexOf('?');
    let primaryPath;
    let queryMap;
    if (searchIndex >= 0) {
        primaryPath = part.slice(0, searchIndex);
        queryMap = (0, _utils_1.parseSearch)(part.slice(searchIndex));
    }
    else {
        primaryPath = part;
        queryMap = new Map();
    }
    const buildingPartGroups = [];
    if (primaryPath) {
        buildingPartGroups.push({
            name: undefined,
            path: primaryPath,
        });
    }
    if (queryMap) {
        for (const group of groups) {
            const key = `_${group}`;
            if (queryMap.has(key)) {
                buildingPartGroups.push({
                    name: group,
                    path: queryMap.get(key),
                });
                queryMap.delete(key);
            }
        }
    }
    return {
        groups: buildingPartGroups,
        query: queryMap,
    };
}
//# sourceMappingURL=route-builder.js.map